/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 2 -*-
 *
 *  Copyright 2005  Carlos Garnacho Parro
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 *  Authors: Carlos Garnacho Parro  <carlosg@gnome.org>
 *
 */

#include "config.h"
#include <gtk/gtk.h>
#include <glib/gi18n-lib.h>

#if EVOLUTION_VERSION < 22900
#include <mail/em-popup.h>
#else
#include <misc/e-popup-action.h>
#include <shell/e-shell-view.h>
#include <shell/es-event.h>
#endif

#include <mail/em-utils.h>

//#include <libedataserver/md5-utils.h>
//#include <e-util/e-error.h>

#if (DATASERVER_VERSION >= 2031001)
#include <camel/camel.h>
#else
#include <camel/camel-folder.h>
#include <camel/camel-medium.h>
#include <camel/camel-stream.h>
#include <camel/camel-stream-mem.h>
#endif

static gint
run_dialog (gint n_duplicates)
{
	gchar *str;
	gint response;

	str = g_strdup_printf ("%i", n_duplicates);
#if EVOLUTION_VERSION < 22900
	response = e_error_run(NULL,
			"org.gnome.remove-duplicates:delete-duplicates",
			str, NULL);
#else
	response = e_alert_run_dialog_for_args(
			NULL,
			"org.gnome.remove-duplicates:delete-duplicates",
			str, NULL);
#endif
	g_free (str);

	return response;
}

static void
delete_message_foreach (gpointer value, gpointer data)
{
	gchar *uid = value;
	CamelFolder *folder = data;

	camel_folder_delete_message (folder, uid);
}

static guchar*
get_message_md5 (CamelFolder *folder,
	const gchar *uid)
{
	CamelMimeMessage *msg;
#if EVOLUTION_VERSION < 23191
	CamelException ex;
#else
	GError *error = NULL;
#endif
	CamelDataWrapper *content;
	CamelStreamMem *mem;

	char res[17], *f;
	guint8 *md5sum;
	gsize length;
	gchar *result;
	int i;
	const char tohex[16] = "0123456789abcdef";
	GChecksum *checksum;
	GByteArray *buffer = g_byte_array_new ();

#if EVOLUTION_VERSION < 23191
	camel_exception_init (&ex);
	msg = camel_folder_get_message (folder, uid, &ex);
	if (camel_exception_is_set (&ex)) {
		camel_exception_clear (&ex);
		return NULL;
	}
#else
#if (DATASERVER_VERSION >= 2033001)
	msg = camel_folder_get_message_sync (folder, uid, NULL, &error);
#else
	msg = camel_folder_get_message (folder, uid, &error);
#endif
	if (error != NULL)
		return NULL;
#endif


	/* get message contents */
	content = camel_medium_get_content ((CamelMedium *) msg);

	if (!content)
		return NULL;

	/* calculate MD5 */
	mem = (CamelStreamMem *)camel_stream_mem_new_with_byte_array(buffer);
#if (DATASERVER_VERSION >= 2033001)
	camel_data_wrapper_decode_to_stream_sync (
		content, (CamelStream *)mem, NULL, NULL);
#else
	camel_data_wrapper_decode_to_stream (
		content, (CamelStream *)mem, NULL);
#endif
	g_byte_array_append (buffer, "\0", 1);

	length = g_checksum_type_get_length (G_CHECKSUM_MD5);
	md5sum = g_alloca (length);

	checksum = g_checksum_new (G_CHECKSUM_MD5);
	g_checksum_update (checksum, (guchar *) buffer->data, -1);
	g_checksum_get_digest (checksum, md5sum, &length);
	g_checksum_free (checksum);

	for (i=0, f = res; i<16;i++) {
		unsigned int c = md5sum[i];
		*f++ = tohex[c & 0xf];
	}
	*f++ = 0;
	result = g_strdup((gchar *)res);

#if (DATASERVER_VERSION >= 2031001)
	g_object_unref (mem);
	g_object_unref (msg);
#else
	camel_object_unref (mem);
	camel_object_unref (msg);
#endif

	return result;
}

static gboolean
message_is_duplicated (GHashTable *messages, guint64 id, guchar *digest)
{
	guchar *hash_digest;
	int i;

	hash_digest = g_hash_table_lookup (messages, &id);

	if (!hash_digest)
		return FALSE;

	for (i = 0; i < 16; i++)
		if (digest[i] != hash_digest[i])
			return FALSE;

	return TRUE;
}

static void
handle_messages (CamelFolder *folder, GPtrArray *uids)
{
	CamelMessageInfo *msg_info;
	const CamelSummaryMessageID *mid;
	GHashTable *messages;
	GSList *duplicates = NULL;
	guint32 flags;
	guchar *digest;
	gint i, n_duplicates;

	messages = g_hash_table_new_full (g_int_hash, g_int_equal, g_free, g_free);

	for (i = 0; i < uids->len; i++) {
		msg_info = camel_folder_get_message_info (folder, uids->pdata[i]);
		digest   = get_message_md5 (folder, uids->pdata[i]);
		mid      = camel_message_info_message_id (msg_info);
		flags    = camel_message_info_flags (msg_info);

		if (!(flags & CAMEL_MESSAGE_DELETED)) {
			if (message_is_duplicated (messages, mid->id.id, digest))
				duplicates = g_slist_prepend (duplicates, g_strdup (uids->pdata[i]));
			else {
				guint64 *id;

				id = g_new0 (guint64, 1);
				*id = mid->id.id;
				g_hash_table_insert (messages, id, digest);
			}
		}

		camel_message_info_free (msg_info);
	}

	n_duplicates = g_slist_length (duplicates);

	if (n_duplicates && (run_dialog (n_duplicates) == GTK_RESPONSE_OK))
		g_slist_foreach (duplicates, delete_message_foreach, folder);

	g_hash_table_destroy (messages);
	g_slist_foreach (duplicates, (GFunc) g_free, NULL);
	g_slist_free (duplicates);
}

void
#if EVOLUTION_VERSION < 22900
org_gnome_duplicates_remove (EPlugin *ep, EMPopupTargetSelect *target);
#else
org_gnome_duplicates_remove (GtkAction *action, EShellView *shell_view);
#endif

void
#if EVOLUTION_VERSION < 22900
org_gnome_duplicates_remove (EPlugin *ep, EMPopupTargetSelect *target)
#else
org_gnome_duplicates_remove (GtkAction *action,  EShellView *shell_view)
#endif
{
	CamelFolder *folder;
	GPtrArray *uids;

#if EVOLUTION_VERSION < 22900
	uids = target->uids;
	folder = target->folder;
#else
	EShellContent *shell_content;
	EMailReader *reader;

	shell_content = e_shell_view_get_shell_content (shell_view);

	reader = E_MAIL_READER (shell_content);
	folder = e_mail_reader_get_folder(reader);
	uids = e_mail_reader_get_selected_uids(reader);
#endif
	handle_messages (folder, uids);
}

static GtkActionEntry entries[] = {
	{ "remove-duplicates",
	"user-trash",
	N_("_Remove Duplicates"),
	NULL,
	N_("Removes duplicated mails from current folder"),
	G_CALLBACK (org_gnome_duplicates_remove) },
};

#if EVOLUTION_VERSION >= 22900
static EPopupActionEntry popup_entries[] = {
	{ "remove-duplicates-popup",
	NULL,
	"remove-duplicates" },
};

gboolean
e_plugin_ui_init (GtkUIManager *ui_manager,
	EShellView *shell_view);

gboolean
e_plugin_ui_init (GtkUIManager *ui_manager,
	EShellView *shell_view)
{
	EShellWindow *shell_window;
	GtkActionGroup *action_group;

	shell_window = e_shell_view_get_shell_window (shell_view);
	action_group = e_shell_window_get_action_group (
			shell_window,
			"mail");

	gtk_action_group_add_actions (
		action_group,
		entries,
		G_N_ELEMENTS (entries),
		shell_view);
	e_action_group_add_popup_actions (
		action_group,
		popup_entries,
		G_N_ELEMENTS (popup_entries));

	return TRUE;
}
#endif

